"""Optional base class for all shared resources used in recipe steps.

A shared resource is something that should persist between single step
actions.  They are useful primarily for not repeating setup operations
that could be time consuming multiple times in a single recipe.

For example, you could setup a shared hardware_manager resource that
exits for the life of your recipe and call many steps that individually
use that hardware_manager to perform actions on an IOTile device.
"""

class SharedResource:
    """Base class for all shared resources."""

    def __init__(self):
        self.opened = False

    def open(self):
        """Open or create this resource.

        This function should be overloaded in subclasses to provide
        the actual implementation needed to open the resource.
        """

        raise NotImplementedError()

    def close(self):
        """Close or destry this resource.

        This function should be overloaded in subclasses to provide
        the actual implementation needed to open the resource.
        """

        raise NotImplementedError()



from .recipe import RecipeObject
from .recipe_manager import RecipeManager
from .exceptions import RecipeFileInvalid, UnknownRecipeResourceType, UnknownRecipeActionType, RecipeVariableNotPassed

__all__ = ['RecipeManager', 'RecipeObject', 'RecipeFileInvalid', 'UnknownRecipeResourceType', 'UnknownRecipeActionType',
           'RecipeVariableNotPassed']



import time
from iotile.ship.exceptions import RecipeActionMissingParameter


class WaitStep:
    """A Recipe Step used to waits a certain amount of time

    WaitStep stops the recipe from running for a set number of seconds

    Args:
        seconds (int): The number of seconds to wait.
    """
    def __init__(self, args):
        if args.get('seconds') is None:
            raise RecipeActionMissingParameter("WaitStep Parameter Missing", \
                parameter_name='seconds')

        self._seconds = args['seconds']

    def run(self):
        time.sleep(float(self._seconds))



import json
import pytest
from iotile.ship.actions.ModifyJsonStep import ModifyJsonStep

class mock_resource:
    def __init__(self, root):
        self.root = root

class test_harness:
    def __init__(self, path):
        self.resources = {'filesystem': mock_resource(path)}
        self.tested_json = path / "tested.json"
        self.expected_json = path / "expected.json"


    def init_data(self, original, expected):
        with open(str(self.tested_json), "w") as file:
            json.dump(original, file, indent=2)

        with open(str(self.expected_json), "w") as file:
            json.dump(expected, file, indent=2)


    def execute_test(self, args):
        step = ModifyJsonStep(args)
        step.run(self.resources)


    def check_success(self, expected):
        with open(str(self.tested_json), "r") as file:
            result = json.load(file)
            assert result == expected

        with open(str(self.tested_json), "r") as file:
            result_str = file.read()

        with open(str(self.expected_json), "r") as file:
            expected_str = file.read()

        assert result_str == expected_str


def test_str(tmp_path):
    harness = test_harness(tmp_path)
    data = {'iotile-id': '42', 'key2' : 'value2'}
    args = {'key': ['iotile-id'],
            'value': 'NEWVAL',
            'path': harness.tested_json}
    expected = {'iotile-id': 'NEWVAL', 'key2' : 'value2'}

    harness.init_data(data, expected)
    harness.execute_test(args)
    harness.check_success(expected)


def test_str_missing(tmp_path):
    harness = test_harness(tmp_path)
    data = {'iotile-id': '42', 'key2' : 'value2'}
    args = {'key': ['NEW'],
            'value': 'NEWVAL',
            'path': harness.tested_json}
    expected = {'iotile-id': '42', 'key2' : 'value2', 'NEW': 'NEWVAL'}

    # Make sure it fails without 'create' flag
    harness.init_data(data, expected)
    with pytest.raises(KeyError):
        harness.execute_test(args)

    # Make sure it works with 'create' flag
    args['create_if_missing'] = True
    harness.execute_test(args)
    harness.check_success(expected)


def test_two_deep(tmp_path):
    harness = test_harness(tmp_path)
    data = {'layer1': {'layer2': {'iotile-id': '42'}}, 'key2' : 'value2'}
    args = {'key': ['layer1', 'layer2', 'iotile-id'],
            'value': 'NEWVAL',
            'path': harness.tested_json}
    expected = {'layer1': {'layer2': {'iotile-id': 'NEWVAL'}}, 'key2' : 'value2'}

    harness.init_data(data, expected)
    harness.execute_test(args)
    harness.check_success(expected)



"""PromptStep. Currently used to stall the recipe prior to running"""
from iotile.ship.exceptions import RecipeActionMissingParameter


class PromptStep:  # pylint:disable=too-few-public-methods; all Steps have one "run" function
    """A Recipe Step used to prompt the user for input.

    Currently, this step doesn't store any information typed into
    the input. It currently stalls the recipe from running before it
    progresses

    Args:
        message (str): The message that will be displayed
    """
    def __init__(self, args):
        if args.get('message') is None:
            raise RecipeActionMissingParameter("PromptStep Parameter Missing", parameter_name='message')

        self._message = args['message']

    def run(self):
        """Get user input to exit the pause"""
        input(str(self._message))



import pytest
from iotile.ship.actions.ModifyJsonStep import modify_dict

def test_str():
    data = {'iotile-id': '42', 'key2' : 'value2'}
    key = ['iotile-id']
    value = 'NEWVAL'
    expected = {'iotile-id': value, 'key2' : 'value2'}

    result = modify_dict(data, key, value)
    assert result == expected

    # Try to modify key that doesn't exist
    key = 'MISSING'
    with pytest.raises(KeyError):
        result = modify_dict(data, key, value)

    # Make sure we're not changing the original
    assert data == {'iotile-id': '42', 'key2' : 'value2'}


def test_str_missing():
    data = {'iotile-id': '42', 'key2' : 'value2',}
    key = ['NEW']
    value = 'NEWVAL'
    expected = {'iotile-id': '42', 'key2' : 'value2', 'NEW': 'NEWVAL'}

    with pytest.raises(KeyError):
        result = modify_dict(data, key, value)

    result = modify_dict(data, key, value, create_if_missing=True)
    assert result == expected


def test_one_deep():
    data = {'layer1': {'iotile-id': '42'}, 'key2' : 'value2'}
    key = ['layer1', 'iotile-id']
    value = 'NEWVAL'
    expected = {'layer1': {'iotile-id': value}, 'key2' : 'value2'}

    result = modify_dict(data, key, value)
    assert result == expected

    # Make sure we're not changing the original
    assert data == {'layer1': {'iotile-id': '42'}, 'key2' : 'value2'}


def test_one_deep_missing():
    data = {'layer1': {'iotile-id': '42'}, 'key2' : 'value2'}
    key = ['layer1', 'NEW']
    value = 'NEWVAL'
    expected = {'layer1': {'iotile-id': '42', 'NEW': 'NEWVAL'}, 'key2' : 'value2'}

    with pytest.raises(KeyError):
        result = modify_dict(data, key, value)

    result = modify_dict(data, key, value, create_if_missing=True)
    assert result == expected


def test_one_deep_missing_modify_first_layer():
    data = {'layer1': {'iotile-id': '42'}, 'key2' : 'value2'}
    key = ['NEW']
    value = 'NEWVAL'
    expected = {'layer1': {'iotile-id': '42'}, 'key2' : 'value2', 'NEW': 'NEWVAL'}

    with pytest.raises(KeyError):
        result = modify_dict(data, key, value)

    result = modify_dict(data, key, value, create_if_missing=True)
    assert result == expected


def test_two_deep():
    data = {'layer1': {'layer2': {'iotile-id': '42'}}, 'key2' : 'value2'}
    key = ['layer1', 'layer2', 'iotile-id']
    value = 'NEWVAL'
    expected = {'layer1': {'layer2': {'iotile-id': value}}, 'key2' : 'value2'}

    result = modify_dict(data, key, value)
    assert result == expected


def test_two_deep_change_middle():
    data = {'layer1': {'iotile-id': '42', 'layer2': {'key_l2' : 'val_l2'}}, 'key2' : 'value2'}
    key = ['layer1', 'iotile-id']
    value = 'NEWVAL'
    expected = {'layer1': {'iotile-id': value, 'layer2': {'key_l2' : 'val_l2'}}, 'key2' : 'value2'}

    result = modify_dict(data, key, value)
    assert result == expected


def test_keychain_includes_a_not_dict():
    data = {'layer1': {'layer2': {'iotile-id': '42'}}, 'key2' : 'NOTADICT'}
    key = ['key2', 'layer2', 'iotile-id']
    value = 'NEWVAL'

    with pytest.raises(ValueError):
        result = modify_dict(data, key, value) # pylint: disable=unused-variable



from iotile.core.hw.update import UpdateScript
from iotile.core.exceptions import ArgumentError


class SendOTAScriptStep:
    """Send a TRUB OTA script to a device and execute it.

    This function requires a shared hardware manager resource to be setup
    containing a connected device that we can send the script to.

    Shared Resources:
        connection (hardware_manager): A connected hardware_manager resource
            that we can use to send our ota script down to a device.

    Args:
        file (str): The path to the target ota file that should be sent.  If
            a relative path is given, it is taken as relative to the current
            working directory.
        no_reboot (bool): Optional argument to not reboot the device
            after the OTA script has been run.  Normal behavior is to
            reboot after OTA.
    """

    REQUIRED_RESOURCES = [('connection', 'hardware_manager')]
    FILES = ['file']

    def __init__(self, args):
        if 'file' not in args:
            raise ArgumentError("SendOTAScriptStep required parameters missing", required=["file"], args=args)

        self._file = args['file']
        self._no_reboot = args.get('no_reboot', False)

        with open(self._file, "rb") as infile:
            data = infile.read()
            self._script = UpdateScript.FromBinary(data)

    def run(self, resources):
        """Actually send the trub script.

        Args:
            resources (dict): A dictionary containing the required resources that
                we needed access to in order to perform this step.
        """

        hwman = resources['connection']

        updater = hwman.hwman.app(name='device_updater')
        updater.run_script(self._script, no_reboot=self._no_reboot)



"""This file contains the schema verifier for verifying a recipe file."""

from iotile.core.utilities.schema_verify import (DictionaryVerifier, Verifier, StringVerifier, IntVerifier, ListVerifier,
                                                 OptionsVerifier, BooleanVerifier)


ActionItem = DictionaryVerifier(desc="A description of a single action that should be run")
ActionItem.add_required("name", StringVerifier("The name of the action type that should be executed"))
ActionItem.add_optional("description", StringVerifier("A short description for what the action is doing"))
ActionItem.add_optional("use", ListVerifier(StringVerifier("The name of a resource"), desc="A list of used resources"))
ActionItem.add_optional("open_before", ListVerifier(StringVerifier("The name of a resource"), desc="A list of resources to open before this step"))
ActionItem.add_optional("close_after", ListVerifier(StringVerifier("The name of a resource"), desc="A list of resources to close after this step"))
ActionItem.key_rule(None, Verifier("A parameter passed into the underlying action"))  # Allow any additional values

ResourceItem = DictionaryVerifier(desc="A shared resource that can be used by one or more action steps")
ResourceItem.add_required("name", StringVerifier("The name of the resource so it can be referenced in a step"))
ResourceItem.add_required("type", StringVerifier("A unique string identifying what type of resource should be created"))
ResourceItem.add_optional("autocreate", BooleanVerifier(desc="Whether to automatically open the resource before the recipe and close it after the recipe"))
ResourceItem.key_rule(None, Verifier("A parameter passed into the underlying resouce"))  # Allow any additional values

VariableDefault = DictionaryVerifier(desc="A default value for a variable", fixed_length=1)
VariableDefault.key_rule(None, StringVerifier())

RecipeSchema = DictionaryVerifier(desc="A recipe containing a list of actions")
RecipeSchema.add_optional("name", StringVerifier("A descriptive name for this recipe"))
RecipeSchema.add_required("description", StringVerifier("A description of what the recipe does"))
RecipeSchema.add_optional("idempotent", BooleanVerifier(desc="Whether the recipe can be run multiple times without breaking"))
RecipeSchema.add_required("actions", ListVerifier(ActionItem.clone(), min_length=1, desc="A list of steps to perform to realize this recipe"))
RecipeSchema.add_optional("resources", ListVerifier(ResourceItem.clone(), desc="An optional list of shared resources to setup"))
RecipeSchema.add_optional("defaults", ListVerifier(VariableDefault.clone(), desc="An optional list of default values for free recipe variables"))



import pathlib

from iotile.core.utilities.schema_verify import DictionaryVerifier
from iotile.core.utilities.schema_verify import StringVerifier
from .shared_resource import SharedResource

RESOURCE_ARG_SCHEMA = DictionaryVerifier(desc="filesystem_manager arguments")
RESOURCE_ARG_SCHEMA.add_required("path", StringVerifier("path to filesystem to work with"))


class FilesystemManagerResource(SharedResource):
    """ A Resource for operating on a Filesystem

    All Steps using this Resource will use self.root as the root directory to find other files

    Root directory is a pathlib.Path

    Arguments:
        path (str): Path to filesystem to work with
    """

    ARG_SCHEMA = RESOURCE_ARG_SCHEMA
    _allowed_types = ["file", "block"]

    def __init__(self, args):
        super(FilesystemManagerResource, self).__init__()

        self._path = args.get('path')
        self.root = None

    def open(self):
        if not pathlib.Path(self._path).exists():
            raise FileNotFoundError("FilesystemManagerResource path not found")

        self.root = pathlib.Path(self._path)
        self.opened = True

    def close(self):
        self.opened = False



import json
import copy
from typedargs import type_system

from iotile.ship.exceptions import RecipeActionMissingParameter

def modify_dict(data, key, value, create_if_missing=False):
    """ Change (or add) a json key/value pair.

    Args:
        data (dict): The original data. This will not be modified.
        key (list): A list of keys and subkeys specifing the key to change (list can be one)
        value (str): The value to change for the above key
        create_if_missing (bool): Set to true to create key if the last key in the list is not found
                Otherwise the function will throw a KeyError
    Returns:
        (dict): the final modified dict
    """
    data_copy = copy.deepcopy(data)
    key_copy = copy.deepcopy(key)

    delver = data_copy
    current_key = key_copy
    last_key = "Root"

    # Dig through the json, setting delver to the dict that contains the last key in "key"
    while len(current_key) > 1:
        if current_key[0] not in delver:
            raise KeyError("ModifyJsonStep Key Couldn't find Subkey {} in {}.".format(current_key[0], last_key))

        if len(current_key) > 2 and not isinstance(delver[current_key[0]], dict):
            raise ValueError("ModifyJsonStep The Value of {} is a {}, not a dict".format(current_key[0], type(delver[current_key[0]])))

        last_key = current_key[0]
        delver = delver[current_key[0]]
        current_key.pop(0)

    if current_key[0] not in delver and not create_if_missing:
        raise KeyError("ModifyJsonStep Key Couldn't find Subkey {} in {}.".format(current_key[0], last_key))

    delver[current_key[0]] = value

    return data_copy

class ModifyJsonStep: # pylint: disable=too-few-public-methods
    """ A Recipe Step used to change (or add) a json key/value pair.

    Args:
        key (list): A list of keys and subkeys specifing the key to change (list can be one)
        value (str): The value to change for the above key
        value_type (str): The type of value. Typedargs types supported. Optional, defaults to 'string'
        path (str): The file path, relative to the Filesystem Manager's root, to the file to change
        create_if_missing (bool): Set to true to create key if the last key in the list is not found
                Otherwise the function will throw a KeyError. Optional, defaults to False.
    """

    REQUIRED_RESOURCES = [('filesystem', 'filesystem_manager')]

    def __init__(self, args):
        if args.get('key') is None:
            raise RecipeActionMissingParameter("ModifyJsonStep Parameter Missing", parameter_name='key')
        if args.get('value') is None:
            raise RecipeActionMissingParameter("ModifyJsonStep Parameter Missing", parameter_name='value')
        if args.get('path') is None:
            raise RecipeActionMissingParameter("ModifyJsonStep Parameter Missing", parameter_name='path')

        self._key = args.get('key')
        self._value_type = str(args.get('value_type', 'string'))
        self._value = type_system.convert_to_type(args.get('value'), self._value_type)
        self._path = str(args.get('path'))
        self._create = type_system.convert_to_type(args.get('create_if_missing', False), 'bool')

    def run(self, resources): # pylint: disable=missing-docstring
        root = resources['filesystem'].root
        fullpath = root.joinpath(self._path)

        with open(str(fullpath), "r") as infile:
            data = json.load(infile)

        result = modify_dict(data, self._key, self._value, self._create)

        with open(str(fullpath), "w") as outfile:
            json.dump(result, outfile, indent=2)



import time
from iotile.core.hw.hwmanager import HardwareManager
from iotile.core.exceptions import ArgumentError


class VerifyDeviceStep:
    """A Recipe Step used to verify that a device is setup as expected

    Args:
        os_tag (int): Optional. Expected os tag to check against in cloud
        os_version (str): Optional. Expected os version to check against in cloud.
            format is "X.Y"
        app_tag (int): Optional. Expected app tag to check against in cloud
        app_version (str): Optional. Expected app version to check against in cloud.
            format is "X.Y"
        realtime_streams (list of int): Optional. List of expected realtime streams variables
            that should be expected upon streaming enabling. If entered in hex, should have 
            the format 0x1XXX. Example: 0x100F
        tile_versions (dict of int: str): Optional. Dictionary of addresses to expection version 
            strings with format "(X, Y, Z)". Example:  8: "(2, 11, 0)"
    """
    REQUIRED_RESOURCES = [('connection', 'hardware_manager')]

    def __init__(self, args):
        self._os_tag            = args.get('os_tag')
        self._os_version        = args.get('os_version')
        self._app_tag           = args.get('app_tag')
        self._app_version       = args.get('app_version')

        self._realtime_streams  = args.get('realtime_streams')
        self._tile_versions     = args.get('tile_versions')

    def _verify_tile_versions(self, hw):
        """Verify that the tiles have the correct versions
        """
        for tile, expected_tile_version in self._tile_versions.items():
            actual_tile_version = str(hw.get(tile).tile_version())
            if expected_tile_version != actual_tile_version:
                raise ArgumentError("Tile has incorrect firmware", tile=tile, \
                    expected_version=expected_tile_version, actual_version=actual_tile_version)

    def _verify_os_app_settings(self, hw):
        """Verify that the os and app tags/versions are set correctly
        """
        con = hw.controller()
        info = con.test_interface().get_info()
        if self._os_tag is not None:
            if info['os_tag'] != self._os_tag:
                raise ArgumentError("Incorrect os_tag", actual_os_tag=info['os_tag'],\
                        expected_os_tag=self._os_tag)
        if self._app_tag is not None:
            if info['app_tag'] != self._app_tag:
                raise ArgumentError("Incorrect app_tag", actual_os_tag=info['app_tag'],\
                        expected_os_tag=self._app_tag)
        if self._os_version is not None:
            if info['os_version'] != self._os_version:
                raise ArgumentError("Incorrect os_version", actual_os_version=info['os_version'],\
                        expected_os_version=self._os_version)
        if self._app_version is not None:
            if info['app_version'] != self._app_version:
                raise ArgumentError("Incorrect app_version", actual_os_version=info['app_version'],\
                        expected_os_version=self._app_version)

    def _verify_realtime_streams(self, hw):
        """Check that the realtime streams are being produced
        """
        print("--> Testing realtime data (takes 2 seconds)")
        time.sleep(2.1)
        reports = [x for x in hw.iter_reports()]
        reports_seen = {key: 0 for key in self._realtime_streams}

        for report in reports:
            stream_value = report.visible_readings[0].stream
            if reports_seen.get(stream_value) is not None:
                reports_seen[stream_value] += 1

        for stream in reports_seen.keys():
            if reports_seen[stream] < 2:
                raise ArgumentError("Realtime Stream not pushing any reports", stream=hex(stream), \
                    reports_seen=reports_seen[stream])

    def run(self, resources):
        hw = resources['connection'].hwman
        if self._tile_versions is not None:
            print('--> Verifying tile versions')
            self._verify_tile_versions(hw)

        if self._os_tag is not None or self._app_tag is not None:
            print('--> Verifying os/app tags and versions')
            self._verify_os_app_settings(hw)

        hw.enable_streaming()

        if self._realtime_streams is not None:
            print('--> Verifying realtime streams')
            self._verify_realtime_streams(hw)



"""Autobuilder plugin for creating a .ship archive from a yaml file.

This file is only meant to be used in conjunction with iotile-build and
should not be imported independently.  It is only for use inside SConstruct
files."""
import os
import imp
import inspect
from SCons.Script import Environment, Copy, Action
from iotile.build.build import ProductResolver, ArchitectureGroup
from iotile.build.utilities import render_template_inplace
from iotile.core.exceptions import BuildError
from ..recipe_manager import RecipeManager

def autobuild_shiparchive(src_file):
    """Create a ship file archive containing a yaml_file and its dependencies.

    If yaml_file depends on any build products as external files, it must
    be a jinja2 template that references the file using the find_product
    filter so that we can figure out where those build products are going
    and create the right dependency graph.

    Args:
        src_file (str): The path to the input yaml file template.  This
            file path must end .yaml.tpl and is rendered into a .yaml
            file and then packaged into a .ship file along with any
            products that are referenced in it.
    """

    if not src_file.endswith('.tpl'):
        raise BuildError("You must pass a .tpl file to autobuild_shiparchive", src_file=src_file)

    env = Environment(tools=[])

    family = ArchitectureGroup('module_settings.json')
    target = family.platform_independent_target()
    resolver = ProductResolver.Create()

    #Parse through build_step products to see what needs to imported
    custom_steps = []
    for build_step in family.tile.find_products('build_step'):
        full_file_name = build_step.split(":")[0]
        basename = os.path.splitext(os.path.basename(full_file_name))[0]
        folder = os.path.dirname(full_file_name)

        fileobj, pathname, description = imp.find_module(basename, [folder])
        mod = imp.load_module(basename, fileobj, pathname, description)
        full_file_name, class_name = build_step.split(":")
        custom_steps.append((class_name, getattr(mod, class_name)))
    env['CUSTOM_STEPS'] = custom_steps

    env["RESOLVER"] = resolver

    base_name, tpl_name = _find_basename(src_file)
    yaml_name = tpl_name[:-4]
    ship_name = yaml_name[:-5] + ".ship"

    output_dir = target.build_dirs()['output']
    build_dir = os.path.join(target.build_dirs()['build'], base_name)
    tpl_path = os.path.join(build_dir, tpl_name)
    yaml_path = os.path.join(build_dir, yaml_name)
    ship_path = os.path.join(build_dir, ship_name)
    output_path = os.path.join(output_dir, ship_name)

    # We want to build up all related files in
    # <build_dir>/<ship archive_folder>/
    # - First copy the template yaml over
    # - Then render the template yaml
    # - Then find all products referenced in the template yaml and copy them
    # - over
    # - Then build a .ship archive
    # - Then copy that archive into output_dir

    ship_deps = [yaml_path]

    env.Command([tpl_path], [src_file], Copy("$TARGET", "$SOURCE"))

    prod_deps = _find_product_dependencies(src_file, resolver)

    env.Command([yaml_path], [tpl_path], action=Action(template_shipfile_action, "Rendering $TARGET"))

    for prod in prod_deps:
        dest_file = os.path.join(build_dir, prod.short_name)
        ship_deps.append(dest_file)
        env.Command([dest_file], [prod.full_path], Copy("$TARGET", "$SOURCE"))

    env.Command([ship_path], [ship_deps], action=Action(create_shipfile, "Archiving Ship Recipe $TARGET"))
    env.Command([output_path], [ship_path], Copy("$TARGET", "$SOURCE"))


def _find_basename(input_file):
    base_src = os.path.basename(input_file[:-4])  # remove .tpl

    base_name, _base_ext = os.path.splitext(base_src)

    return base_name, base_src + ".tpl"


def _find_product_dependencies(src_file, resolver):
    resolver.start_tracking()
    render_template_inplace(src_file, {}, dry_run=True, resolver=resolver)

    products = resolver.end_tracking()
    return products


def template_shipfile_action(target, source, env):
    """Template a yaml ship recipe in place."""

    render_template_inplace(str(source[0]), {}, resolver=env["RESOLVER"])


def create_shipfile(target, source, env):
    """Create a .ship file with all dependencies."""

    source_dir = os.path.dirname(str(source[0]))
    recipe_name = os.path.basename(str(source[0]))[:-5]

    resman = RecipeManager()

    resman.add_recipe_actions(env['CUSTOM_STEPS'])
    resman.add_recipe_folder(source_dir, whitelist=[os.path.basename(str(source[0]))])
    recipe = resman.get_recipe(recipe_name)

    recipe.archive(str(target[0]))



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

# Caveats and possible issues
# Mac OS X
# - when using a virtualenv, readline is not properly installed into the virtualenv
#   and cannot be imported.  You need to install it using easy_install as described here
#   http://calvinx.com/tag/readline/


from setuptools import setup, find_packages
import version

setup(
    name="iotile-ship",
    packages=find_packages(exclude=("test",)),
    version=version.version,
    license="LGPLv3",
    install_requires=[
        "iotile-core>=5.0.0,<6",
        "pyaml>=18.11.0,<19",
        "pyyaml>5.1,<6"
    ],
    python_requires=">=3.5,<4",
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'iotile-ship = iotile.ship.scripts.iotile_ship:main'
        ],
        'iotile.recipe_action': [
            'PromptStep         = iotile.ship.actions.prompt_step:PromptStep',
            'WaitStep           = iotile.ship.actions.wait_step:WaitStep',
            'PipeSnippetStep    = iotile.ship.actions.pipe_snippet_step:PipeSnippetStep',
            'SyncCloudStep      = iotile.ship.actions.sync_cloud_step:SyncCloudStep',
            'SyncRTCStep        = iotile.ship.actions.sync_rtc_step:SyncRTCStep',
            'VerifyDeviceStep   = iotile.ship.actions.verify_device_step:VerifyDeviceStep',
            'SendOTAScriptStep  = iotile.ship.actions.send_ota_script_step:SendOTAScriptStep',
            'ModifyJsonStep     = iotile.ship.actions.ModifyJsonStep:ModifyJsonStep'
        ],
        'iotile.recipe_resource': [
            'hardware_manager   = iotile.ship.resources:HardwareManagerResource',
            'filesystem_manager = iotile.ship.resources:FilesystemManagerResource'
        ],
        'iotile.autobuild': [
            'autobuild_shiparchive = iotile.ship.autobuild:autobuild_shiparchive'
        ]
    },
    description="IOTile Ship Support",
    author="Arch",
    author_email="info@arch-iot.com",
    url="http://github.com/iotile/coretools/iotileship",
    keywords=["iotile", "arch", "embedded", "hardware", "firmware"],
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries :: Python Modules"
        ],
    long_description="""\
IOTileship
-----------

A python package for shiping embedded firmware on IOTile based devices.  See https://www.arch-iot.com.

IOTileship embeds SCons in accordance with its license permitting redistribution.  More information on
SCons can be found at: https://scons.org
"""
)



import os
from iotile.core.dev import ComponentRegistry
from iotile.ship.recipe import RecipeObject
from iotile.ship.exceptions import RecipeNotFoundError


class RecipeManager:
    """A class that maintains a list of installed recipes and recipe actions.

    It allows fetching recipes by name and auotmatically building RecipeObjects
    from textual descriptions.

    The RecipeManager maintains a dictionary of RecipeAction objects that it
    compiles from all installed iotile packages. It passes this dictionary to
    any Recipe that is created from it so the recipe can find any recipe
    actions that it needs.

    The RecipeManager finds RecipeActions by looking for plugins that
    are registered with pkg_resources.
    """
    def __init__(self):

        self._recipe_actions = {}
        self._recipe_resources = {}
        self._recipes = {}

        reg = ComponentRegistry()
        for name, action in reg.load_extensions('iotile.recipe_action', product_name='build_step'):
            self._recipe_actions[name] = action

        for name, resource in reg.load_extensions('iotile.recipe_resource'):
            self._recipe_resources[name] = resource

    def is_valid_action(self, name):
        """Check if a name describes a valid action.

        Args:
            name (str): The name of the action to check

        Returns:
            bool: Whether the action is known and valid.
        """

        return self._recipe_actions.get(name, None) is not None

    def is_valid_recipe(self, recipe_name):
        """Check if a recipe is known and valid.

        Args:
            name (str): The name of the recipe to check

        Returns:
            bool: Whether the recipe is known and valid.
        """

        return self._recipes.get(recipe_name, None) is not None

    def add_recipe_folder(self, recipe_folder, whitelist=None):
        """Add all recipes inside a folder to this RecipeManager with an optional whitelist.

        Args:
            recipe_folder (str): The path to the folder of recipes to add.
            whitelist (list): Only include files whose os.basename() matches something
                on the whitelist
        """

        if whitelist is not None:
            whitelist = set(whitelist)

        if recipe_folder == '':
            recipe_folder = '.'

        for yaml_file in [x for x in os.listdir(recipe_folder) if x.endswith('.yaml')]:
            if whitelist is not None and yaml_file not in whitelist:
                continue

            recipe = RecipeObject.FromFile(os.path.join(recipe_folder, yaml_file), self._recipe_actions, self._recipe_resources)
            self._recipes[recipe.name] = recipe

        for ship_file in [x for x in os.listdir(recipe_folder) if x.endswith('.ship')]:
            if whitelist is not None and ship_file not in whitelist:
                continue

            recipe = RecipeObject.FromArchive(os.path.join(recipe_folder, ship_file), self._recipe_actions, self._recipe_resources)
            self._recipes[recipe.name] = recipe

    def add_recipe_actions(self, recipe_actions):
        """Add additional valid recipe actions to RecipeManager

        args:
            recipe_actions (list): List of tuples. First value of tuple is the classname,
                second value of tuple is RecipeAction Object

        """
        for action_name, action in recipe_actions:
            self._recipe_actions[action_name] = action

    def get_recipe(self, recipe_name):
        """Get a recipe by name.

        Args:
            recipe_name (str): The name of the recipe to fetch. Can be either the
                yaml file name or the name of the recipe.
        """
        if recipe_name.endswith('.yaml'):
            recipe = self._recipes.get(RecipeObject.FromFile(recipe_name, self._recipe_actions, self._recipe_resources).name)
        else:
            recipe = self._recipes.get(recipe_name)
        if recipe is None:
            raise RecipeNotFoundError("Could not find recipe", recipe_name=recipe_name, known_recipes=[x for x in self._recipes.keys()])

        return recipe




from iotile.core.utilities.schema_verify import DictionaryVerifier, OptionsVerifier, StringVerifier, IntVerifier
from iotile.core.hw import HardwareManager
from iotile.core.exceptions import HardwareError
from .shared_resource import SharedResource


RESOURCE_ARG_SCHEMA = DictionaryVerifier(desc="hardware_manager arguments")
RESOURCE_ARG_SCHEMA.add_optional("port", StringVerifier("the port string to use to connect to devices"))
RESOURCE_ARG_SCHEMA.add_optional("connect", OptionsVerifier(StringVerifier(), IntVerifier(), desc="the uuid of the device to connect to"))
RESOURCE_ARG_SCHEMA.add_optional("connect_direct", OptionsVerifier(StringVerifier(), desc="the connection string of the device to connect to"))


class HardwareManagerResource(SharedResource):
    """A shared HardwareManager instance.

    Arguments:
        port (str): Optional port argument that should
            be used to connect to a DeviceAdapter.  If not
            specified, the virtual environment default is used.
        connect: (str or int): The UUID of a device to connect to
            when this resource is created and disconnect from when
            it is destroyed.  This is an optional parameter.  If
            it is not specified the HardwareManager is not connected
            upon creation.
        connect_direct: (str or int): The connection string of a device to connect
            directly to when this resource is created and disconnect from
            when it is destroyed.  This is an optional parameter.  If
            it is not specified the HardwareManager is not connected
            upon creation.
    """

    ARG_SCHEMA = RESOURCE_ARG_SCHEMA

    def __init__(self, args):
        super(HardwareManagerResource, self).__init__()

        self._port = args.get('port')
        self._connect_id = args.get('connect')
        self._connection_string = args.get('connect_direct')
        self.hwman = None

        if self._connect_id is not None and not isinstance(self._connect_id, int):
            self._connect_id = int(self._connect_id, 0)

    def open(self):
        """Open and potentially connect to a device."""

        self.hwman = HardwareManager(port=self._port)
        self.opened = True

        if self._connection_string is not None:
            try:
                self.hwman.connect_direct(self._connection_string)
            except HardwareError:
                self.hwman.close()
                raise

        elif self._connect_id is not None:
            try:
                self.hwman.connect(self._connect_id)
            except HardwareError:
                self.hwman.close()
                raise


    def close(self):
        """Close and potentially disconnect from a device."""

        if self.hwman.stream.connected:
            self.hwman.disconnect()

        self.hwman.close()
        self.opened = False
